// ============================================================
//  BouncingBall.cpp	1996 Hiroshi Lockheimer
// ============================================================

#include "BouncingBall.h"


CBouncingBallThread	*gBouncer = NULL;


void
module_initialize(
	void	*inSettings,
	long 	inSettingsSize)
{
#pragma unused (inSettings, inSettingsSize)
}


void
module_cleanup(
	void	**outSettings,
	long	*outSettingsSize)
{
	*outSettings = NULL;
	*outSettingsSize = 0;
}


void
module_start_saving(
	BView	*inView)
{	
	srand((long)real_time_clock());

	inView->Window()->Show();

	gBouncer = new CBouncingBallThread();
	gBouncer->StartSaving(inView);	
}


void
module_stop_saving()
{
	gBouncer->StopSaving();
	delete (gBouncer);
	gBouncer = NULL;
}


void
module_start_config(
	BView *inView)
{
	if (!inView->Window()->Lock())
		return;

	BStringView *view = NULL;
	
	view = new BStringView(BRect(20.0, 20.0, 100.0, 32.0),
								 B_EMPTY_STRING,
								 "BouncingBall");
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());
	view->SetFontName("Emily");
	
	view = new BStringView(BRect(30.0, 40.0, 130.0, 52.0),
								 B_EMPTY_STRING,
								 "by Hiroshi Lockheimer");
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());
	
	inView->Window()->Unlock();
}


void
module_stop_config()
{
}


CBouncingBallThread::CBouncingBallThread()
	: CSaveThread()
{
}


void
CBouncingBallThread::StartSaving(
	BView	*view)
{
	mBallList.AddItem(new CBouncingBall(view));
	mBallList.AddItem(new CBouncingBall(view));
		
	CSaveThread::StartSaving(view);
}

	
void
CBouncingBallThread::StopSaving()
{
	CSaveThread::StopSaving();
	
	CBouncingBall *ball = NULL;
	for ( long i = 0; 
		  (ball = (CBouncingBall *)mBallList.ItemAt(i)) != NULL; 
		  i++)
		delete (ball);
}

	
double
CBouncingBallThread::Save()
{
	CBouncingBall *ball = NULL;
	for ( long i = 0; 
		  (ball = (CBouncingBall *)mBallList.ItemAt(i)) != NULL; 
		  i++)
	ball->NextStep();
		
	return (10000.0);
}


const double	kBallSpeed = 25.0;
const long		kBounceCount = 10;


CBouncingBall::CBouncingBall(
	BView		*inView)
{	
	mView = inView;
	mWindow = mView->Window();
	
	if (!mWindow->Lock())
		return;

	// set up a random initial position of ball
	mXDir = mYDir = 1;	// sign of velocities is positive
	mBallDiameter = 30.0;
	
	// place the ball at a random location in the view
	BRect viewRect = mView->Bounds();
	mBallRect.left = RangedRandom(viewRect.left, viewRect.right - mBallDiameter);
	mBallRect.right = mBallRect.left + mBallDiameter;
	mBallRect.top = RangedRandom(viewRect.top, viewRect.bottom - mBallDiameter);
	mBallRect.bottom = mBallRect.top + mBallDiameter;

	// initialize counter that keeps track of number of bounces that have occurred
	mBounceCount = kBounceCount;
		
	// choose a random direction for the ball to travel
	double ballAngle = RandomAngle(10, 80);
	mBallSin = sin(ballAngle);
	mBallCos = cos(ballAngle);
	
	// pick a color
	mBallColor = system_colors()->color_list[RangedRandom(1, 255)];

	mWindow->Unlock();
}


void
CBouncingBall::NextStep()
{
	bool bouncing = FALSE;
	
	if (!mWindow->Lock())
		return;
	
	// calculate what the current x & y components of velocity should be
	short	xVelocity = kBallSpeed * mBallCos;
	short	yVelocity = kBallSpeed * mBallSin;

	// move the ball
	mBallRect.OffsetBy(mXDir * xVelocity, mYDir * yVelocity);
	
	// now that we've moved the ball, see if it is still visible
	// if not, the ball should bounce off the wall and change direction
	BRect viewRect = mView->Bounds();
	if (!viewRect.Contains(mBallRect)) {
		short deltaX = 0;
		short deltaY = 0;
		
		// we are going to bounce the ball
		bouncing = TRUE;
		
		// see what directions we have to change
		if (mBallRect.right > viewRect.right) {
			mXDir = -1;					// we're off to the right
			deltaX = viewRect.right - mBallRect.right;
		}
		if (mBallRect.left < viewRect.left) {
			mXDir = 1;					// we're off to the left
			deltaX = viewRect.left - mBallRect.left;
		}
		if (mBallRect.bottom > viewRect.bottom) {
			mYDir = -1;					// we're off below
			deltaY = viewRect.bottom - mBallRect.bottom;
		}
		if (mBallRect.top < viewRect.top) {
			mYDir = 1;					// we're off above
			deltaY = viewRect.top - mBallRect.top;
		}
		
		// make the ball appear to hit the wall right at its boundary
		mBallRect.OffsetBy(deltaX, deltaY);
	}
	
	// are we bouncing?
	if ((bouncing) && (mBounceCount-- == 0)) {
		// time to choose a new angle
		// choose an angle between 10 and 80 
		double ballAngle = RandomAngle(10, 80);
		mBounceCount = kBounceCount;
		mBallSin = sin(ballAngle);
		mBallCos = cos(ballAngle);
	}

	// draw it!
	mView->SetHighColor(mView->LowColor());
	mView->FillRoundRect(mOldRect, 15.0, 15.0);
	
	mView->SetHighColor(mBallColor);
	mView->FillRoundRect(mBallRect, 15.0, 15.0);

	mOldRect = mBallRect;
	
	mWindow->Unlock();
}


short
RangedRandom(
	short	lower,
	short	upper)
{
	short middle = (lower + upper) / 2;
	short randValue = middle + (rand() % (upper - middle));
	if(randValue < lower)
		return (lower);
	if(randValue > upper)
		return (upper);
		
	return (randValue);
}


double
RandomAngle(
	short	lower,
	short	upper)
{
	double randomDouble = RangedRandom(lower, upper);
	return ((randomDouble * M_PI_4) / 45.0);
}